from matplotlib.contour import *
import _mpl_tri

class TriContourSet(ContourSet):
    """
    Underlying implementation of triangular grid contouring functions.

    Modified version of matplotlib's ContourSet; much is common to ContourSet.
    """
    def __init__(self, ax, *args, **kwargs):
        self.ax = ax
        self.levels = kwargs.get('levels', None)
        self.filled = kwargs.get('filled', False)
        self.linewidths = kwargs.get('linewidths', None)
        self.linestyles = kwargs.get('linestyles', 'solid')

        self.alpha = kwargs.get('alpha', 1.0)
        self.origin = kwargs.get('origin', None)
        self.extent = kwargs.get('extent', None)
        cmap = kwargs.get('cmap', None)
        self.colors = kwargs.get('colors', None)
        norm = kwargs.get('norm', None)
        self.extend = kwargs.get('extend', 'neither')
        self.antialiased = kwargs.get('antialiased', True)
        self.nchunk = kwargs.get('nchunk', 0)
        self.locator = kwargs.get('locator', None)
        if (isinstance(norm, colors.LogNorm)
            or isinstance(self.locator, ticker.LogLocator)):
            self.logscale = True
            if norm is None:
                norm = colors.LogNorm()
            if self.extend is not 'neither':
                raise ValueError('extend kwarg does not work yet with log scale')
        else:
            self.logscale = False

        if self.origin is not None: assert(self.origin in
                                           ['lower', 'upper', 'image'])
        if self.extent is not None: assert(len(self.extent) == 4)
        if cmap is not None: assert(isinstance(cmap, colors.Colormap))
        if self.colors is not None and cmap is not None:
            raise ValueError('Either colors or cmap must be None')
        if self.origin == 'image': self.origin = mpl.rcParams['image.origin']
        x, y, z, triangles, mask = self._contour_args(*args, **kwargs)
        if self.colors is not None:
            cmap = colors.ListedColormap(self.colors, N=len(self.layers))
        if self.filled:
            self.collections = cbook.silent_list('collections.PolyCollection')
        else:
            self.collections = cbook.silent_list('collections.LineCollection')
        # label lists must be initialized here
        self.labelTexts = []
        self.labelCValues = []

        kw = {'cmap': cmap}
        if norm is not None:
            kw['norm'] = norm
        cm.ScalarMappable.__init__(self, **kw) # sets self.cmap;
        self._process_colors()
        _mask = ma.getmask(z)
        if _mask is ma.nomask:
            _mask = None

        if self.filled:
            if self.linewidths is not None:
                warnings.warn('linewidths is ignored by contourf')
            triGrid = _mpl_tri.PyTriGrid(x, y, triangles, mask)
            lowers = self._levels[:-1]
            uppers = self._levels[1:]
            for lower, upper in zip(lowers, uppers):
                contourLines = triGrid.contour(z, lower, upper)
                #print contourLines
                col = collections.PolyCollection(contourLines,
                                                 antialiaseds=(self.antialiased,),
                                                 edgecolors= 'none',
                                                 alpha=self.alpha)
                self.ax.add_collection(col, False)
                self.collections.append(col)
                #col2 = collections.LineCollection(contourLines, linewidths=[2])
                #self.ax.add_collection(col2, False)
                #self.collections.append(col2)
        else:
            tlinewidths = self._process_linewidths()
            self.tlinewidths = tlinewidths
            tlinestyles = self._process_linestyles()

            triGrid = _mpl_tri.PyTriGrid(x, y, triangles, mask)
            for level,width,lstyle in zip(self.levels,tlinewidths,tlinestyles):
                contourLines = triGrid.contour(z, level)
                #print contourLines
                col = collections.LineCollection(contourLines,
                                                 linewidths = width,
                                                 linestyle = lstyle,
                                                 alpha = self.alpha)
                if level < 0.0 and self.monochrome:
                    ls = mpl.rcParams['contour.negative_linestyle']
                    col.set_linestyle(ls)
                col.set_label('_nolegend_')
                self.ax.add_collection(col, False)
                self.collections.append(col)
        self.changed() # set the colors
        x0 = ma.minimum(x)
        x1 = ma.maximum(x)
        y0 = ma.minimum(y)
        y1 = ma.maximum(y)
        self.ax.update_datalim([(x0,y0), (x1,y1)])
        self.ax.autoscale_view()

    def _contour_args(self, *args, **kwargs):
        if self.filled: fn = 'contourf'
        else:           fn = 'contour'
        Nargs = len(args)

        if Nargs < 4:
            raise TypeError("Insufficient arguments to %s" % fn)
        elif Nargs > 5:
            raise TypeError("Too many arguments to %s" % fn)

        x = args[0]
        y = args[1]
        triangles = args[2]
        z = args[3]

        # This should maybe in equivalent of self._check_xyz
       	x = np.asarray(x, dtype=np.float)
	y = np.asarray(y, dtype=np.float)
	if x.shape != y.shape or len(x.shape) != 1:
            raise ValueError('x and y must be equal-length 1D arrays')
	
	triangles = np.asarray(triangles, dtype=np.short)
	if len(triangles.shape) != 2 or triangles.shape[1] != 3:
            raise ValueError('triangles must be a (?,3) array')
	if triangles.max() >= len(x):
            raise ValueError('triangles max element is out of bounds')
	if triangles.min() < 0:
            raise ValueError('triangles min element is out of bounds')

	z = np.asarray(z)
	if z.shape != x.shape or len(z.shape) != 1:
            raise ValueError('z must be same length 1D array as x and y')

	# Optional mask.
	mask = kwargs.get('mask', None)
	if mask != None:
            kwargs.pop('mask')
            mask = np.asarray(mask, dtype=np.bool)
            if len(mask.shape) != 1 or mask.shape[0] != triangles.shape[0]:
                raise ValueError('mask and triangles arrays must have same '+
                                 'length')

        self.zmax = ma.maximum(z)
        self.zmin = ma.minimum(z)
        if self.logscale and self.zmin <= 0:
            z = ma.masked_where(z <= 0, z)
            warnings.warn('Log scale: values of z <=0 have been masked')
            self.zmin = z.min()
        self._auto = False
        if self.levels is None:
            if Nargs == 4:
                lev = self._autolev(z, 7)
            else:   # 5 args
                level_arg = args[-1]
                try:
                    if type(level_arg) == int:
                        lev = self._autolev(z, level_arg)
                    else:
                        lev = np.asarray(level_arg).astype(np.float64)
                except:
                    raise TypeError(
                        "Last %s arg must give levels; see help(%s)" % (fn,fn))
            if self.filled and len(lev) < 2:
                raise ValueError("Filled contours require at least 2 levels.")
            # Workaround for cntr.c bug wrt masked interior regions:
            #if filled:
            #    z = ma.masked_array(z.filled(-1e38))
            # It's not clear this is any better than the original bug.
            self.levels = lev
        #if self._auto and self.extend in ('both', 'min', 'max'):
        #    raise TypeError("Auto level selection is inconsistent "
        #                             + "with use of 'extend' kwarg")
        self._levels = list(self.levels)
        if self.extend in ('both', 'min'):
            self._levels.insert(0, min(self.levels[0],self.zmin) - 1)
        if self.extend in ('both', 'max'):
            self._levels.append(max(self.levels[-1],self.zmax) + 1)
        self._levels = np.asarray(self._levels)
        self.vmin = np.amin(self.levels)  # alternative would be self.layers
        self.vmax = np.amax(self.levels)
        if self.extend in ('both', 'min'):
            self.vmin = 2 * self.levels[0] - self.levels[1]
        if self.extend in ('both', 'max'):
            self.vmax = 2 * self.levels[-1] - self.levels[-2]
        self.layers = self._levels # contour: a line is a thin layer
        if self.filled:
            self.layers = 0.5 * (self._levels[:-1] + self._levels[1:])
            if self.extend in ('both', 'min'):
                self.layers[0] = 0.5 * (self.vmin + self._levels[1])
            if self.extend in ('both', 'max'):
                self.layers[-1] = 0.5 * (self.vmax + self._levels[-2])

        return (x, y, z, triangles, mask)
