import matplotlib.pyplot as lab
import numpy as np

def triplot(x, y, triangles, **kwargs):
    """
    Draw triangular grid as lines and/or points.

    Arguments:
      x, y - arrays of point x and y-coordinates with shape (npoints).
      triangles - array of triangle point indices with shape (ntriangles,3).

    Keyword arguments:
      mask - optional boolean mask array with shape (ntriangles).

      Other keyword arguments are passed to underlying matplotlib plot
      function, e.g. 'c' for color, 'marker' for marker type, etc.

    Limitations:
      Points are not rendered correctly for alpha < 1, but lines are.
    """
    x = np.asarray(x)
    y = np.asarray(y)
    if x.shape != y.shape or len(x.shape) != 1:
        raise ValueError('x and y must be equal-length 1D arrays')
	
    triangles = np.asarray(triangles)
    if len(triangles.shape) != 2 or triangles.shape[1] != 3:
        raise ValueError('triangles must be a (?,3) array')
    if triangles.max() >= len(x):
        raise ValueError('triangles max element is out of bounds')
    if triangles.min() < 0:
        raise ValueError('triangles min element is out of bounds')

    # Optional mask.
    mask = kwargs.get('mask', None)
    if mask != None:
        kwargs.pop('mask')
        mask = np.asarray(mask, dtype=np.bool)
        if len(mask.shape) != 1 or mask.shape[0] != triangles.shape[0]:
            raise ValueError('mask and triangles arrays must have same length')

    # Unnecessary calculations if only plotting points.
    # Each point can be drawn more than once, so is incorrect if alpha < 1.
    # Should really separate drawing of markers and lines.

    # Apply triangle mask.
    if mask != None:
        triangles = triangles.compress(1-mask, axis=0)

    # Array (?,2) of all triangle edges.
    edges = np.vstack((triangles[:,[0,1]],
                       triangles[:,[1,2]],
                       triangles[:,[2,0]]))

    # Sort each edge so that lowest point index is first.
    edges.sort()

    # Sort edges list so that lowest indices are first.
    edges = edges.tolist()
    edges.sort()

    # Remove duplicates from sorted list (python cookbook).
    edges = [edge for i,edge in enumerate(edges) if not i or edge != edges[i-1]]
    edges = np.asarray(edges)

    lab.plot(x[edges].T, y[edges].T, **kwargs)
