#include "HoleRemover.h"
#include <algorithm>
#include <cassert>
#include <map>

HoleRemover::HoleRemover()
{}

HoleRemover::~HoleRemover()
{}

int HoleRemover::insertHoleIntoContainer(ContourLine& target,
					 int targetEdge,
					 const ContourLine& source,
					 int sourceEdge,
					 const XY& point)
{
  unsigned int startSize = target.size();

  contourLineInsertUniquePoint(target, target.begin() + targetEdge, point);
  target.insert(target.begin() + targetEdge,
		source.begin(), source.begin() + sourceEdge + 1);
  target.insert(target.begin() + targetEdge,
		source.begin() + sourceEdge, source.end());
  contourLineInsertUniquePoint(target, target.begin() + targetEdge, point);
 
  return target.size() - startSize;
}

bool HoleRemover::intersectsBelowPoint(const XY& point,
				       const XY& start,
				       const XY& end,
				       double& y) const
{
  if (start.x != end.x)
  {
    double m = (end.y - start.y) / (end.x - start.x);
    double c = start.y - m*start.x;
    y = m*point.x + c;
    if (y <= point.y)
    {
      return ( (start.x >= point.x && end.x <= point.x) ||
	       (start.x <= point.x && end.x >= point.x) );
    }
  }
  else if (start.x == point.x)   // point, start and end all have same x.
  {
    if (start.y <= point.y && end.y <= point.y)
    {
      // Both start and end below y.
      y = max(start.y, end.y);
      return true;
    }
    else if ( (start.y <= point.y && end.y >= point.y) ||
	      (start.y >= point.y && end.y <= point.y) )
    {
      // One of each of start,end is above,below point.
      y = point.y;
      return true;
    }
  }
  return false;
}

void HoleRemover::removeHoles(Contour& contour)
{
  // For each contour line, calculate lowest point, whether a hole or not, and
  // bounding box.
  unsigned int nLines = contour.size();
  vector<LineItem> lineItems(nLines);
  for (unsigned int i = 0; i < nLines; ++i)
  {
    const ContourLine& line = contour[i];
    int lowestEdge = min_element(line.begin(), line.end()) - line.begin();
    const XY& lowestPoint = line[lowestEdge];

    lineItems[i].lowestEdge = lowestEdge;
    lineItems[i].lowestPoint = lowestPoint;

    XY next = line[(lowestEdge + 1) % line.size()];
    XY prev = line[(lowestEdge + line.size() - 1) % line.size()];
    double angleDiff = (next-lowestPoint).angle() - (lowestPoint-prev).angle();
    while (angleDiff > 2*M_PI)
    {
      angleDiff -= 2*M_PI;
    }
    while (angleDiff < 0)
    {
      angleDiff += 2*M_PI;
    }
    lineItems[i].isHole = (angleDiff > M_PI);

    for (unsigned int j = 0; j < line.size(); ++j)
    {
      lineItems[i].bBox.add(line[j]);
    }
  }

  /*
  cout << "lowest points" << endl;
  for (unsigned int i = 0; i < nLines; ++i)
  {
    cout << "  " << i << ": " << lineItems[i].lowestEdge << " "
	 << lineItems[i].lowestPoint << ", bbox " << lineItems[i].bBox
	 << ", hole=" << lineItems[i].isHole << endl;
  }
  */

  // For holes, find line,edge immediately below lowest point.
  typedef map<int, HoleItem> HoleMap;
  HoleMap holeMap;
  for (unsigned int i = 0; i < nLines; ++i)
  {
    if (lineItems[i].isHole)
    {
      const XY& point = lineItems[i].lowestPoint;
      int targetLine = -1;
      int targetEdge = -1;
      double nearestY = 0;
      double y = 0;
      for (unsigned i2 = 0; i2 < nLines; ++i2)
      {
	if (i2 != i &&
	    lineItems[i2].bBox.containsX(point.x) &&
	    lineItems[i2].bBox.lower.y <= point.y &&
	    (targetLine == -1 || lineItems[i2].bBox.upper.y > nearestY))
	{
	  const ContourLine& line = contour[i2];
	  unsigned int nPoints = line.size();
	  for (unsigned int j = 0; j < nPoints; ++j)
	  {
	    if (intersectsBelowPoint(point, line[j], line[(j+1)%nPoints], y))
	    {
	      if (targetEdge == -1 || y > nearestY)
	      {
		targetLine = i2;
		targetEdge = j;
		nearestY = y;
	      }
	    }
	  }
	}
      }

      assert(targetLine != -1);
      HoleItem item(targetLine, targetEdge, XY(point.x, nearestY));
      holeMap.insert(HoleMap::value_type(i, item));
      lineItems[targetLine].containingHoleCount++;
    }
  }

  /*
  cout << "holeMap" << endl;
  for (HoleMap::const_iterator it = holeMap.begin(); it != holeMap.end(); ++it)
  {
    cout << "  " << it->first << " -> " << it->second.targetLine << " "
	 << it->second.targetEdge << " " << it->second.targetPoint << endl;
  }
  cout << "containingHoles" << endl;
  for (unsigned int i = 0; i < nLines; ++i)
  {
    cout << "  " << i << ": " << lineItems[i].containingHoleCount << endl;
  }
  */

  // Insert holes into containing lines.
  while (!holeMap.empty())
  {
    for (HoleMap::iterator it = holeMap.begin(); it != holeMap.end(); ++it)
    {
      int sourceLine = it->first;
      if (lineItems[sourceLine].containingHoleCount == 0)
      {
	int targetLine = it->second.targetLine;
	ContourLine& target = contour[targetLine];
	int targetEdge = (it->second.targetEdge + 1) % target.size();
	ContourLine& source = contour[sourceLine];
	int sourceEdge = lineItems[sourceLine].lowestEdge;
	
	int pointsAdded = insertHoleIntoContainer(target,
						  targetEdge,
						  source,
						  sourceEdge,
						  it->second.targetPoint);
	
	holeMap.erase(it);
	lineItems[targetLine].containingHoleCount--;
	source.erase(source.begin(), source.end());

	// Correct indices on target if they are >= targetEdge.
	if (lineItems[targetLine].lowestEdge >= targetEdge)
	{
	  lineItems[targetLine].lowestEdge += pointsAdded;
	}
	for (HoleMap::iterator it1 = holeMap.begin();
	     it1 != holeMap.end(); ++it1)
	{
	  if (it1->second.targetLine == targetLine &&
	      it1->second.targetEdge >= targetEdge)
	  {
	    it1->second.targetEdge += pointsAdded;
	  }
	}
      }
    }
  }

  // Remove empty contour lines, in reverse order.
  for (int i = (int)nLines-1; i >= 0; --i)
  {
    if (contour[i].empty())
    {
      contour.erase(contour.begin() + i);
    }
  }
}
