#include "TriGrid.h"
#include <cassert>
#include <set>

TriGrid::TriGrid(int nPoints,
		 int nTriangles,
		 const double* x,
		 const double* y,
		 int* trianglePointIndices,
		 const bool* triangleMask)
  : _nPoints(nPoints),
    _nTriangles(nTriangles),
    _x(x),
    _y(y),
    _trianglePointIndices(trianglePointIndices),
    _triangleMask(triangleMask),
    _triangleNeighbours(0)
{
  correctTriangles();
  calculateTriangleNeighbours();
  calculateBoundaries();
}

TriGrid::~TriGrid()
{
  delete [] _triangleNeighbours;
}

void TriGrid::calculateBoundaries()
{
  assert(_boundaries.empty() && "Boundaries already calculated");

  int tri, edge, point;

  // Create set of all boundary edges, which are triangle edges that do not
  // have a neighbour.
  typedef set<TriEdge> BoundaryEdges;
  BoundaryEdges boundaryEdges;
  for (tri = 0; tri < _nTriangles; ++tri)
  {
    if (!getMask(tri))
    {
      for (edge = 0; edge < 3; ++edge)
      {
	if (getTriangleNeighbour(tri, edge) == -1)
	{
	  boundaryEdges.insert(TriEdge(tri, edge));
	}
      }
    }
  }

  /*
  cout << "TriGrid boundaryEdges" << endl;
  for (BoundaryEdges::const_iterator it = boundaryEdges.begin();
       it != boundaryEdges.end(); ++it)
  {
    cout << "  " << it->triangle << " " << it->edge << " = "
	 << getTrianglePoint(*it) << endl;
  }
  */

  // Start with a boundary edge, then use triangle neighbours to determine
  // which is the next boundary edge and so on until return to start.
  while (!boundaryEdges.empty())
  {
    _boundaries.push_back(Boundary());
    Boundary& boundary = _boundaries.back();

    // First edge of boundary.
    BoundaryEdges::iterator it = boundaryEdges.begin();
    tri = it->triangle;
    edge = it->edge;

    while (true)
    {
      _triEdgeToBoundaryMap[TriEdge(tri,edge)] = IntPair(_boundaries.size()-1,
							 boundary.size());
      boundary.push_back(*it);
      boundaryEdges.erase(it);

      // Move to next edge of this triangle.
      edge = (edge+1) % 3;

      // Find next triEdge by traversing valid neighbours.
      while (getTriangleNeighbour(tri, edge) != -1)
      {
	point = getTrianglePointIndex(tri, edge);
	tri = getTriangleNeighbour(tri, edge);
	edge = getEdgeInTriangle(tri, point);
      }

      if (TriEdge(tri,edge) == boundary.front())
      {
	break;
      }
      else
      {
	it = boundaryEdges.find(TriEdge(tri, edge));
	assert(it != boundaryEdges.end() && "Incorrect boundary edge");
      }
    }
  }

  //writeBoundaries(cout);
}

void TriGrid::calculateTriangleNeighbours()
{
  assert(_triangleNeighbours == 0 && "Triangle neighbours already calculated");

  int tri, edge;

  // Map of edges to the two triangles that have those edges, or -1 if no such
  // triangle.  Edge defined by start and end point indices, start < end.
  typedef map<IntPair, IntPair> EdgeToTrianglesMap;
  EdgeToTrianglesMap edgeToTrianglesMap;

  for (tri = 0; tri < _nTriangles; ++tri)
  {
    if (!getMask(tri))
    {
      for (edge = 0; edge < 3; ++edge)
      {
	int start = getTrianglePointIndex(tri, edge);
	int end   = getTrianglePointIndex(tri, (edge+1)%3);
	bool startLower = (start < end);
	IntPair se(startLower ? start : end, startLower ? end : start);
	EdgeToTrianglesMap::iterator it = edgeToTrianglesMap.find(se);
	if (it == edgeToTrianglesMap.end())
	{
	  it = edgeToTrianglesMap.insert(EdgeToTrianglesMap::value_type(
	         se, IntPair(-1,-1))).first;
	}
	if (startLower)
	{
	  it->second.first = tri;
	}
	else
	{
	  it->second.second = tri;
	}
      }
    }
  }

  /*
  for (EdgeToTrianglesMap::const_iterator it = edgeToTrianglesMap.begin();
       it != edgeToTrianglesMap.end(); ++it)
  {
    cout << it->first.first << ' ' << it->first.second << " -> "
	 << it->second.first << ' ' << it->second.second << endl;
  }
  */

  // Calculate triangle neighbours.
  _triangleNeighbours = new int[_nTriangles*3];
  int i = 0;
  for (tri = 0; tri < _nTriangles; ++tri)
  {
    if (getMask(tri))
    {
      _triangleNeighbours[i++] = -1;
      _triangleNeighbours[i++] = -1;
      _triangleNeighbours[i++] = -1;
    }
    else
    {
      for (edge = 0; edge < 3; ++edge)
      {
	int start = getTrianglePointIndex(tri, edge);
	int end   = getTrianglePointIndex(tri, (edge+1)%3);
	bool startLower = (start < end);
	EdgeToTrianglesMap::const_iterator it =
	  edgeToTrianglesMap.find(IntPair(startLower ? start : end,
					  startLower ? end : start));
	assert(it != edgeToTrianglesMap.end() &&
               "Incorrect edge to triangles map");
	_triangleNeighbours[i++] =
	  (startLower ? it->second.second : it->second.first);
      }
    }
  }

  //writeTriangleNeighbours(cout);
}

void TriGrid::correctTriangles()
{
  for (int i = 0; i < 3*_nTriangles; i += 3)
  {
    XY point0 = getPoint(_trianglePointIndices[i]);
    if ( (getPoint(_trianglePointIndices[i+1]) - point0).crossZ(
	    getPoint(_trianglePointIndices[i+2])-point0) < 0.0)
    {
      // Triangle points are clockwise, so change them to anticlockwise.
      swap(_trianglePointIndices[i+1], _trianglePointIndices[i+2]);
    } 
  }
}

const TriGrid::Boundaries& TriGrid::getBoundaries() const
{
  return _boundaries;
}

int TriGrid::getEdgeInTriangle(int triangle, int point) const
{
  assert(triangle >= 0 && triangle < _nTriangles &&
	 "Triangle index out of bounds.");
  assert(point >= 0 && point < _nPoints && "Point index out of bounds.");

  for (int edge = 0; edge < 3; ++edge)
  {
    if (getTrianglePointIndex(triangle, edge) == point)
    {
      return edge;
    }
  }
  assert(0 && "Triangle does not contain point");
  return 0;
}

bool TriGrid::getMask(int triangle) const
{
  assert(triangle >= 0 && triangle < _nTriangles &&
	 "Triangle index out of bounds.");
  return _triangleMask != 0 && _triangleMask[triangle];
}

XY TriGrid::getPoint(int pointIndex) const
{
  assert(pointIndex >= 0 && pointIndex < _nPoints &&
	 "Point index out of bounds");
  return XY(_x[pointIndex], _y[pointIndex]);
}

int TriGrid::getTriangleCount() const
{
  return _nTriangles;
}

int TriGrid::getTriangleNeighbour(int triangle, int edge) const
{
  assert(triangle >= 0 && triangle < _nTriangles &&
	 "Triangle index out of bounds");
  assert(edge >= 0 && edge < 3 && "Edge index out of bounds");
  return _triangleNeighbours[triangle*3 + edge];
}

TriEdge TriGrid::getTriangleNeighbourEdge(const TriEdge& triEdge) const
{
  int tri = getTriangleNeighbour(triEdge.triangle, triEdge.edge);
  if (tri == -1)
  {
    return TriEdge(-1,-1);
  }
  else
  {
    return TriEdge(
      tri, getEdgeInTriangle(tri, getTrianglePointIndex(triEdge.triangle,
							(triEdge.edge+1)%3)));
  }
}

XY TriGrid::getTrianglePoint(int triangle, int edge) const
{
  return getPoint(getTrianglePointIndex(triangle, edge));
}

XY TriGrid::getTrianglePoint(const TriEdge& triEdge) const
{
  return getPoint(getTrianglePointIndex(triEdge));
}

int TriGrid::getTrianglePointIndex(int triangle, int edge) const
{
  assert(triangle >= 0 && triangle < _nTriangles &&
	 "Triangle index out of bounds");
  assert(edge >= 0 && edge < 3 && "Edge index out of bounds");
  return _trianglePointIndices[triangle*3 + edge];
}

int TriGrid::getTrianglePointIndex(const TriEdge& triEdge) const
{
  return getTrianglePointIndex(triEdge.triangle, triEdge.edge);
}

const double& TriGrid::getX(int point) const
{
  assert(point >= 0 && point < _nPoints && "Point index out of bounds");
  return _x[point];
}

const double& TriGrid::getY(int point) const
{
  assert(point >= 0 && point < _nPoints && "Point index out of bounds");
  return _y[point];
}

void TriGrid::triEdgeToBoundary(const TriEdge& triEdge,
				int& iBoundary,
				int& iSegment) const
{
  TriEdgeToBoundaryMap::const_iterator it = _triEdgeToBoundaryMap.find(triEdge);
  assert(it != _triEdgeToBoundaryMap.end() && "TriEdge is not on a boundary");
  iBoundary = it->second.first;
  iSegment = it->second.second;
}

void TriGrid::writeBoundaries(ostream& os) const
{
  os << "TriGrid boundaries: " << _boundaries.size() << endl;
  for (Boundaries::const_iterator it = _boundaries.begin();
       it != _boundaries.end(); ++it)
  {
    const Boundary& boundary = *it;
    os << "  Boundary of " << boundary.size() << " points:";
    for (Boundary::const_iterator p = boundary.begin();
	 p != boundary.end(); ++p)
    {
      os << ' ' << getTrianglePoint(p->triangle, p->edge);
    }
    os << endl;
  }
}

void TriGrid::writeTriangleNeighbours(ostream& os) const
{
  os << "TriGrid triangle neighbours" << endl;
  for (int tri = 0; tri < _nTriangles; ++tri)
  {
    os << "  " << tri << ": " << getTriangleNeighbour(tri,0) << " "
	 << getTriangleNeighbour(tri,1) << " " << getTriangleNeighbour(tri,2)
	 << endl;
  }
}
