#ifndef TRI_GRID_H
#define TRI_GRID_H

#include "TriEdge.h"
#include "XY.h"
#include <map>
#include <vector>
using namespace std;

/**
 * Triangular grid.
 */
class TriGrid
{
public:
  /**
   * Typedef for a single boundary loop of a TriGrid.
   */
  typedef vector<TriEdge> Boundary;

  /**
   * Typedef for a collection of boundary loops of a TriGrid.
   */
  typedef vector<Boundary> Boundaries;

  /**
   * Constructor.
   * @param nPoints Number of points.
   * @param nTriangles Number of triangles.
   * @param x Array of x point coordinates of length nPoints.
   * @param y Array of y point coordinates of length nPoints.
   * @param trianglePointIndices Array of triangles point indices of length
   * 3*nTriangles.
   * @param triangleMask Optional triangle mask, either 0 or an array of
   * length nTriangles.
   */
  TriGrid(int nPoints,
	  int nTriangles,
	  const double* x,
	  const double* y,
	  int* trianglePointIndices,
	  const bool* triangleMask);

  /**
   * Destructor.
   */
  ~TriGrid();

  /**
   * Return the collection of boundary loops of this TriGrid.
   */
  const Boundaries& getBoundaries() const;

  /**
   * Return an indication of whether the specified triangle is masked or not.
   */
  bool getMask(int triangle) const;

  /**
   * Return the specified point.
   */
  XY getPoint(int pointIndex) const;

  int getTriangleCount() const;

  /**
   * Return the triangle edge that is the neighbour of the specified triangle
   * edge. Return (-1,-1) if no such neighbour.
   */
  TriEdge getTriangleNeighbourEdge(const TriEdge& triEdge) const;

  XY getTrianglePoint(int triangle, int edge) const;

  XY getTrianglePoint(const TriEdge& triEdge) const;

  int getTrianglePointIndex(int triangle, int edge) const;

  int getTrianglePointIndex(const TriEdge& triEdge) const;

  const double& getX(int point) const;

  const double& getY(int point) const;

  /**
   * Return which boundary and boundary segment the specified triangle edge is.
   */
  void triEdgeToBoundary(const TriEdge& triEdge,
			 int& iBoundary,
			 int& iSegment) const;

  void writeBoundaries(ostream& os) const;

  void writeTriangleNeighbours(ostream& os) const;

private:
  void calculateBoundaries();

  void calculateTriangleNeighbours();

  /**
   * Reorder point indices of triangles so that each is anticlockwise.
   */
  void correctTriangles();

  /**
   * Return the triangle edge index in the range 0 to 2 that the specified
   * point is.
   */
  int getEdgeInTriangle(int triangle, int point) const;

  int getTriangleNeighbour(int triangle, int edge) const;




  int _nPoints;
  int _nTriangles;
  const double* _x;               // Owned by python.
  const double* _y;               // Owned by python.
  int* _trianglePointIndices;     // Owned by python, ordered anticlockwise.
  const bool* _triangleMask;      // Owned by python, may be 0.

  int* _triangleNeighbours;       // Owned.
  Boundaries _boundaries;

  typedef pair<int, int> IntPair;
  typedef map<TriEdge, IntPair> TriEdgeToBoundaryMap;
  TriEdgeToBoundaryMap _triEdgeToBoundaryMap;
};

#endif // TRI_GRID_H
