#include "TriGrid.h"
#include "TriGridContourGenerator.h"
#include <algorithm>
#include <cassert>
#include <iostream>


#define EDGE_INTERP(tri,edge) interp(_triGrid.getTrianglePointIndex(tri,edge), _triGrid.getTrianglePointIndex(tri,(edge+1)%3), level)


TriGridContourGenerator::TriGridContourGenerator(const TriGrid& triGrid)
  : _triGrid(triGrid),
    _interiorVisited(2*triGrid.getTriangleCount()),
    _boundariesVisited(0),
    _boundariesUsed(0)
{}

TriGridContourGenerator::~TriGridContourGenerator()
{}

void TriGridContourGenerator::clearVisited(bool includeBoundaries)
{
  // Clear _interiorVisited.
  fill(_interiorVisited.begin(), _interiorVisited.end(), false);

  if (includeBoundaries)
  {
    if (_boundariesVisited.empty())
    {
      // Initialise _boundariesVisited.
      const TriGrid::Boundaries& boundaries = _triGrid.getBoundaries();
      _boundariesVisited.reserve(boundaries.size());
      for (TriGrid::Boundaries::const_iterator it = boundaries.begin();
	   it != boundaries.end(); ++it)
      {
	_boundariesVisited.push_back(BoundaryVisited(it->size()));
      }

      // Initialise _boundariesUsed.
      _boundariesUsed = BoundariesUsed(boundaries.size());
    }
    
    // Clear _boundariesVisited.
    for (BoundariesVisited::iterator it = _boundariesVisited.begin();
	 it != _boundariesVisited.end(); ++it)
    {
      fill(it->begin(), it->end(), false);
    }

    // Clear _boundariesUsed.
    fill(_boundariesUsed.begin(), _boundariesUsed.end(), false);
  }

  /*
  cout << "BoundariesVisited " << _boundariesVisited.size() << endl;
  for (BoundariesVisited::const_iterator i1 = _boundariesVisited.begin();
       i1 != _boundariesVisited.end(); ++i1)
  {
    cout << " ";
    for (BoundaryVisited::const_iterator i2 = i1->begin(); i2 != i1->end();++i2)
    {
      cout << " " << *i2;
    }
    cout << endl;
  }
  */
}

Contour* TriGridContourGenerator::createContour(const double* z,
						const double& level)
{
  clearVisited(false);
  _z = z;

  Contour* contour = new Contour;

  findBoundaryLines(*contour, level);
  findInteriorLines(*contour, level, false, true);

  return contour;
}

Contour* TriGridContourGenerator::createFilledContour(const double* z,
						      const double& lowerLevel,
						      const double& upperLevel)
{
  clearVisited(true);
  _z = z;

  Contour* contour = new Contour;

  findBoundaryLines2(*contour, lowerLevel, upperLevel);
  findInteriorLines(*contour, lowerLevel, false, false);
  findInteriorLines(*contour, upperLevel, true, false);

  return contour;
}

void TriGridContourGenerator::findBoundaryLines(Contour& contour,
						const double& level)
{
  // Iterate over all boundary edges to find starting points for all contour
  // lines that intersect the boundaries.
  const TriGrid::Boundaries& boundaries = _triGrid.getBoundaries();
  for (TriGrid::Boundaries::const_iterator it = boundaries.begin();
       it != boundaries.end(); ++it)
  {
    const TriGrid::Boundary& boundary = *it;
    bool startAbove, endAbove = false;
    for (TriGrid::Boundary::const_iterator itb = boundary.begin();
	 itb != boundary.end(); ++itb)
    {
      if (itb == boundary.begin())
      {
	startAbove = _z[_triGrid.getTrianglePointIndex(*itb)] >= level;
      }
      else
      {
	startAbove = endAbove;
      }
      endAbove = _z[_triGrid.getTrianglePointIndex(itb->triangle,
						   (itb->edge+1)%3)] >= level;
      if (startAbove && !endAbove)
      {
	// Start point for contour line.
	contour.push_back(ContourLine());
	ContourLine& contourLine = contour.back();
	TriEdge triEdge = *itb;
	followInterior(contourLine, triEdge, true, level, false);
      }
    }
  }
}

void TriGridContourGenerator::findBoundaryLines2(Contour& contour,
						 const double& lowerLevel,
						 const double& upperLevel)
{
  double zStart, zEnd;
  const TriGrid::Boundaries& boundaries = _triGrid.getBoundaries();
  for (unsigned int i = 0; i < boundaries.size(); ++i)
  {
    const TriGrid::Boundary& boundary = boundaries[i];
    for (unsigned j = 0; j < boundary.size(); ++j)
    {
      if (!_boundariesVisited[i][j])
      {
	zStart = _z[_triGrid.getTrianglePointIndex(boundary[j])];
	zEnd = _z[_triGrid.getTrianglePointIndex(boundary[j].triangle,
						 (boundary[j].edge+1)%3)];

	bool incrThroughUpper = (zStart < upperLevel && zEnd >= upperLevel);
	bool decrThroughLower = (zStart >= lowerLevel && zEnd < lowerLevel);

	if (decrThroughLower || incrThroughUpper)
	{
	  // Start point for contour line.
	  contour.push_back(ContourLine());
	  ContourLine& contourLine = contour.back();
	  TriEdge startTriEdge = boundary[j];
	  TriEdge triEdge = startTriEdge;

	  // Traverse interior and boundaries until back at startTriEdge.
	  bool onUpper = incrThroughUpper;
	  do
	  {
	    followInterior(contourLine, triEdge, true,
			   onUpper ? upperLevel : lowerLevel, onUpper);
	    onUpper = followBoundary(contourLine, triEdge, lowerLevel,
				     upperLevel, onUpper);
	  } while (triEdge != startTriEdge);
	}
      }
    }
  }

  /*
  cout << "BoundariesUsed" << endl;
  for (unsigned int i = 0; i < boundaries.size(); ++i)
  {
    cout << "  " << _boundariesUsed[i] << " "
	 << _z[TRIANGLE_POINT(boundaries[i][0].triangle,
			      boundaries[i][0].edge)] << endl;
  }
  */

  // Add full boundaries.
  for (unsigned int i = 0; i < boundaries.size(); ++i)
  {
    if (!_boundariesUsed[i])
    {
      const TriGrid::Boundary& boundary = boundaries[i];
      double z = _z[_triGrid.getTrianglePointIndex(boundary[0])];
      if (z >= lowerLevel && z < upperLevel)
      {
	contour.push_back(ContourLine(boundary.size()));
	ContourLine& contourLine = contour.back();
	for (unsigned int j = 0; j < boundary.size(); ++j)
	{
	  contourLine[j] =
	    _triGrid.getPoint(_triGrid.getTrianglePointIndex(boundary[j]));
	}
      }
    }
  }
}

void TriGridContourGenerator::findInteriorLines(Contour& contour,
						const double& level,
						bool onUpper,
						bool closeLoops)
{
  for (int tri = 0; tri < _triGrid.getTriangleCount(); ++tri)
  {
    int visitedIndex = tri;
    if (onUpper)
    {
      visitedIndex += _triGrid.getTriangleCount();
    }

    if (_interiorVisited[visitedIndex] || _triGrid.getMask(tri))
    {
      // Triangle has already been visited or is masked.
      continue;
    }

    _interiorVisited[visitedIndex] = true;

    // Determine edge via which to leave this triangle.
    int edge = getExitEdge(tri, level, onUpper);
    assert(edge >= -1 && edge < 3 && "Invalid exit edge");
    if (edge == -1)
    {
      // Contour does not pass through this triangle.
      continue;
    }

    // Found start of new line loop.
    contour.push_back(ContourLine());
    ContourLine& contourLine = contour.back();
    TriEdge triEdge = _triGrid.getTriangleNeighbourEdge(TriEdge(tri,edge));
    followInterior(contourLine, triEdge, false, level, onUpper);

    if (closeLoops)
    {
      contourLine.push_back(contourLine.front());
    }
  }
}

bool TriGridContourGenerator::followBoundary(ContourLine& contourLine,
					     TriEdge& triEdge,
					     const double& lowerLevel,
					     const double& upperLevel,
					     bool onUpper)
{
  // Have TriEdge to start at, need equivalent boundary iBoundary and iSegment.
  int iBoundary, iSegment;
  _triGrid.triEdgeToBoundary(triEdge, iBoundary, iSegment);

  _boundariesUsed[iBoundary] = true;

  bool stop = false;
  bool firstSegment = true;
  double zStart, zEnd = 0;
  while (!stop)
  {
    assert(!_boundariesVisited[iBoundary][iSegment] &&
	   "Boundary already visited");
    
    _boundariesVisited[iBoundary][iSegment] = true;
    
    if (firstSegment)
    {
      zStart = _z[_triGrid.getTrianglePointIndex(triEdge)];
    }
    else
    {
      zStart = zEnd;
    }
    zEnd = _z[_triGrid.getTrianglePointIndex(triEdge.triangle,
					     (triEdge.edge+1)%3)];

    if (zEnd > zStart)  // z increasing.
    {
      if (!(!onUpper && firstSegment) &&
	  zEnd >= lowerLevel && zStart < lowerLevel)
      {
	stop = true;
	onUpper = false;
      }
      else if (zEnd >= upperLevel && zStart < upperLevel)
      {
	stop = true;
	onUpper = true;
      }
    }
    else  // z decreasing.
    {
      if (!(onUpper && firstSegment) &&
	  zStart >= upperLevel && zEnd < upperLevel)
      {
	stop = true;
	onUpper = true;
      }
      else if (zStart >= lowerLevel && zEnd < lowerLevel)
      {
	stop = true;
	onUpper = false;
      }
    }

    firstSegment = false;

    if (!stop)
    {
      iSegment++;
      if (iSegment == (int)_triGrid.getBoundaries()[iBoundary].size())
      {
	iSegment = 0;
      }
      triEdge = _triGrid.getBoundaries()[iBoundary][iSegment];

      contourLineAddUniquePoint(
	contourLine,_triGrid.getPoint(_triGrid.getTrianglePointIndex(triEdge)));
    }
  }

  /*
  cout << "BoundariesVisited " << _boundariesVisited.size() << endl;
  for (BoundariesVisited::const_iterator i1 = _boundariesVisited.begin();
       i1 != _boundariesVisited.end(); ++i1)
  {
    cout << " ";
    for (BoundaryVisited::const_iterator i2 = i1->begin(); i2 != i1->end();++i2)
    {
      cout << " " << *i2;
    }
    cout << endl;
  }
  */

  return onUpper;
}

void TriGridContourGenerator::followInterior(ContourLine& contourLine,
					     TriEdge& triEdge,
					     bool endOnBoundary,
					     const double& level,
					     bool onUpper)
{
  int& tri = triEdge.triangle;
  int& edge = triEdge.edge;

  // Initial point.
  contourLineAddUniquePoint(contourLine, EDGE_INTERP(tri,edge));

  while(true)
  {
    int visitedIndex = tri;
    if (onUpper)
    {
      visitedIndex += _triGrid.getTriangleCount();
    }

    // Check for end not on boundary.
    if (!endOnBoundary && _interiorVisited[visitedIndex])
    {
      // Reached start point, so return.
      break;
    }

    // Determine edge by which to leave this triangle.
    edge = getExitEdge(tri, level, onUpper);
    assert(edge >= 0 && edge < 3 && "Invalid exit edge");

    _interiorVisited[visitedIndex] = true;

    // Append new point to point set.
    assert(edge >= 0 && edge < 3 && "Invalid triangle edge");
    contourLineAddUniquePoint(contourLine, EDGE_INTERP(tri,edge));
 
    // Move to next triangle.
    TriEdge nextTriEdge = _triGrid.getTriangleNeighbourEdge(triEdge);

    // Check if ending on a boundary.
    if (endOnBoundary && nextTriEdge.triangle == -1)
    {
      break;
    }

    triEdge = nextTriEdge;
    assert(triEdge.triangle != -1 && "Invalid triangle for internal loop");
  }
}

int TriGridContourGenerator::getExitEdge(int tri,
					 const double& level,
					 bool onUpper) const
{
  unsigned int config =
    (_z[_triGrid.getTrianglePointIndex(tri,0)] >= level) |
    (_z[_triGrid.getTrianglePointIndex(tri,1)] >= level) << 1 |
    (_z[_triGrid.getTrianglePointIndex(tri,2)] >= level) << 2;

  if (onUpper)
  {
    switch (config)
    {
      case 0: return -1;
      case 1: return  0;
      case 2: return  1;
      case 3: return  1;
      case 4: return  2;
      case 5: return  0;
      case 6: return  2;
      case 7: return -1;
      default: assert(0 && "Invalid config value"); return -1;
    }
  }
  else
  {
    switch (config)
    {
      case 0: return -1;
      case 1: return  2;
      case 2: return  0;
      case 3: return  2;
      case 4: return  1;
      case 5: return  1;
      case 6: return  0;
      case 7: return -1;
      default: assert(0 && "Invalid config value"); break;
    }
  }
}

XY TriGridContourGenerator::interp(int point1,
				   int point2,
				   const double& level) const
{
  assert(point1 != point2 && "Identical points");
  double fraction = (_z[point2] - level) / (_z[point2] - _z[point1]);
  return XY(_triGrid.getX(point1)*fraction +
	    _triGrid.getX(point2)*(1.0 - fraction),
	    _triGrid.getY(point1)*fraction +
	    _triGrid.getY(point2)*(1.0 - fraction));
}
