#ifndef TRI_GRID_CONTOUR_GENERATOR_H
#define TRI_GRID_CONTOUR_GENERATOR_H

#include "Contour.h"
#include "TriEdge.h"

// Forward declarations.
class TriGrid;

/**
 * Creator of contours on a particular TriGrid.
 */
class TriGridContourGenerator
{
public:
  TriGridContourGenerator(const TriGrid& triGrid);

  ~TriGridContourGenerator();

  /**
   * Create and return a new contour at a particular level. Caller is
   * responsible for deleting the returned contour.
   */
  Contour* createContour(const double* z,
			 const double& level);

  /**
   * Create and return a new filled contour between two levels. Caller is
   * responsible for deleting the returned contour.
   */
  Contour* createFilledContour(const double* z,
			       const double& lowerLevel,
			       const double& upperLevel);

private:
  /**
   * Clear visited flags.
   */
  void clearVisited(bool includeBoundaries);

  /**
   * Find and follow lines at the specified level that start and end on a
   * boundary of the TriGrid.
   */
  void findBoundaryLines(Contour& contour,
			 const double& level);

  /**
   * Find and follow lines at either of the specified levels that start and
   * end of a boundary of the TriGrid.
   */
  void findBoundaryLines2(Contour& contour,
			  const double& lowerLevel,
			  const double& upperLevel);

  /**
   * Find and follow lines at the specified level that do not intersect a 
   * boundary of the TriGrid.
   */
  void findInteriorLines(Contour& contour,
			 const double& level,
			 bool onUpper,
			 bool closeLoops);

  /**
   * Follow the boundary of the TriGrid from the specified TriEdge to its end
   * which can be on either the lower or upper levels. Only used for filled
   * contours.
   * @return Returns true if on upper level, false if lower.
   */
  bool followBoundary(ContourLine& contourLine,
		      TriEdge& triEdge,
		      const double& lowerLevel,
		      const double& upperLevel,
		      bool onUpper);

  /**
   * switchDirection - whether to switch direction to follow lines. Set to 
   * true if following upperLevel.
   */
  void followInterior(ContourLine& contourLine,
		      TriEdge& triEdge,
		      bool endOnBoundary,
		      const double& level,
		      bool onUpper);

  /**
   * Return the edge by which the specified level leaves the specified
   * triangle, which is 0, 1 or 2 if the contour passes through the triangle
   * or -1 otherwise.
   */
  int getExitEdge(int tri, const double& level, bool onUpper) const;

  /**
   * Interpolate point on contour between the two specified point indices.
   */
  XY interp(int point1, int point2, const double& level) const;





  typedef vector<bool> BoundaryVisited;
  typedef vector<BoundaryVisited> BoundariesVisited;

  typedef vector<bool> BoundariesUsed;

  const TriGrid& _triGrid;
  vector<bool> _interiorVisited;
  BoundariesVisited _boundariesVisited;  // Only used for filled contours.
  BoundariesUsed _boundariesUsed;        // Only used for filled contours.

  // The following are changed each time contour() is called.
  const double* _z;                      // Owned by python.
};

#endif // TRI_GRID_CONTOUR_GENERATOR_H
