#include <Python.h>
#include <numpy/noprefix.h>

#include "Contour.h"
#include "HoleRemover.h"
#include "TriGrid.h"
#include "TriGridContourGenerator.h"

typedef struct
{
  PyObject_HEAD;
  PyArrayObject* xArr;
  PyArrayObject* yArr;
  PyArrayObject* triArr;
  PyArrayObject* maskArr;
  TriGrid* triGrid;
} PyTriGrid;

extern "C"
{
  static int PyTriGrid_clear(PyTriGrid* self)
  {
    Py_XDECREF(self->xArr);
    self->xArr = 0;

    Py_XDECREF(self->yArr);
    self->yArr = 0;

    Py_XDECREF(self->triArr);
    self->triArr = 0;

    Py_XDECREF(self->maskArr);
    self->maskArr = 0;

    delete self->triGrid;
    self->triGrid = 0;
    return 0;
  }

  static PyObject* PyTriGrid_contour(PyTriGrid* self, PyObject* args)
  {
    PyObject* zArg;
    double lowerLevel, upperLevel = -1e100;
    if (!PyArg_ParseTuple(args, "Od|d", &zArg, &lowerLevel, &upperLevel))
    {
      PyErr_SetString(PyExc_ValueError, "Expecting array and 1 or 2 doubles");
      return 0;
    }

    if (!PyArray_Check(zArg))
    {
      PyErr_SetString(PyExc_ValueError, "Expecting z array");
      return 0;
    }

    PyArrayObject* zArr =
      (PyArrayObject*)PyArray_ContiguousFromObject(zArg, NPY_DOUBLE, 1, 1);  
    if (zArr == 0 || zArr->dimensions[0] != self->xArr->dimensions[0])
    {
      PyErr_SetString(PyExc_ValueError,
		      "Argument z must be the same shape as x and y");
      Py_XDECREF(zArr);
      return 0;
    }

    // Create contour.
    TriGridContourGenerator contourGenerator(*self->triGrid);
    Contour* contour = 0;
    if (upperLevel == -1e100 || upperLevel <= lowerLevel)
    {
      contour = contourGenerator.createContour((double*)zArr->data,
					       lowerLevel);
    }
    else
    {
      contour = contourGenerator.createFilledContour((double*)zArr->data,
						     lowerLevel,
						     upperLevel);
      if (contour->size() > 1)
      {
	HoleRemover().removeHoles(*contour);
      }
    }

    Py_XDECREF(zArr);

    // Marshall contour from C++ to python. Note C++ and python have separate
    // copies of the contour data.
    PyObject* pyContour = PyList_New(contour->size());
    for (unsigned int i = 0; i < contour->size(); ++i)
    {
      const ContourLine& line = (*contour)[i];
      int nPoints = line.size();
      npy_intp dims[2] = {nPoints,2};
      PyArrayObject* pyLine = (PyArrayObject*)PyArray_SimpleNew(2, dims,
								NPY_DOUBLE);
      memcpy((double*)pyLine->data, &line[0], 2*nPoints*sizeof(double));
      PyList_SetItem(pyContour, i, (PyObject*)pyLine);
    }

    delete contour;
    return Py_BuildValue("O", pyContour);
  }

  static void PyTriGrid_dealloc(PyTriGrid* self)
  {
    PyTriGrid_clear(self);
    self->ob_type->tp_free((PyObject*)self);
  }

  static int PyTriGrid_init(PyTriGrid* self, PyObject* args)
  {
    PyObject* xArg;
    PyObject* yArg;
    PyObject* triArg;
    PyObject* maskArg;
    if (!PyArg_ParseTuple(args, "OOOO", &xArg, &yArg, &triArg, &maskArg))
    {
      PyErr_SetString(PyExc_ValueError, "Expecting 4 objects");
      return -1;
    }

    if (!PyArray_Check(xArg) || !PyArray_Check(yArg) || !PyArray_Check(triArg))
    {
      PyErr_SetString(PyExc_ValueError, "Expecting x, y and triangles arrays");
      return -1;
    }

    PyArrayObject* xArr =
      (PyArrayObject*)PyArray_ContiguousFromObject(xArg, NPY_DOUBLE, 1, 1);
    PyArrayObject* yArr =
      (PyArrayObject*)PyArray_ContiguousFromObject(yArg, NPY_DOUBLE, 1, 1);
    if (xArr == 0 || yArr == 0 || xArr->dimensions[0] != yArr->dimensions[0])
    {
      PyErr_SetString(PyExc_ValueError,
		      "Arguments x, y must be a 1D arrays of the same length");
      Py_XDECREF(xArr);
      Py_XDECREF(yArr);
      return -1;
    }

    PyArrayObject* triArr =
      (PyArrayObject*)PyArray_ContiguousFromObject(triArg, NPY_INT, 2, 2);
    if (triArr == 0 || triArr->dimensions[1] != 3)
    {
      PyErr_SetString(PyExc_ValueError,
		      "Argument triangles must be a shape (?,3) array of "
		      "integers");
      Py_XDECREF(xArr);
      Py_XDECREF(yArr);
      Py_XDECREF(triArr);
      return -1;
    }

    PyArrayObject* maskArr = 0;
    if (maskArg && maskArg != Py_None)
    {
      maskArr = (PyArrayObject*)PyArray_ContiguousFromObject(maskArg,
							     NPY_BOOL, 1, 1);
      if (maskArr == 0 || maskArr->dimensions[0] != triArr->dimensions[0])
      {
	PyErr_SetString(PyExc_ValueError,
			"Argument mask must be 1D array with same length as "
			"triangles array");
	Py_XDECREF(xArr);
	Py_XDECREF(yArr);
	Py_XDECREF(triArr);
	Py_XDECREF(maskArr);
	return -1;
      }
    }

    self->xArr = xArr;
    self->yArr = yArr;
    self->triArr = triArr;
    self->maskArr = maskArr;
    self->triGrid = new TriGrid((int)xArr->dimensions[0],
				(int)triArr->dimensions[0],
				(const double*)xArr->data,
				(const double*)yArr->data,
				(int*)triArr->data,
				maskArr ? (const bool*)maskArr->data : 0);
    return 0;
  }

  static PyObject* PyTriGrid_new(PyTypeObject* type,
				 PyObject* args,
				 PyObject* kwds)
  {
    PyTriGrid* self = (PyTriGrid*)type->tp_alloc(type, 0);
    if (self != 0)
    {
      self->xArr = 0;
      self->yArr = 0;
      self->triArr = 0;
      self->maskArr = 0;
      self->triGrid = 0;
    }
    return (PyObject*)self;
  }

  static PyMethodDef PyTriGrid_methods[] =
    {
      {"contour", (PyCFunction)PyTriGrid_contour, METH_VARARGS,
       "pydoc da da da d ad ada"},
      {0, 0, 0, 0}
    };

  static PyTypeObject PyTriGridType = {
    PyObject_HEAD_INIT(0)
    0,                               // ob_size
    "mpl_tri.PyTriGrid",             // tp_name
    sizeof(PyTriGrid),               // tp_basicsize
    0,                               // tp_itemsize
    (destructor)PyTriGrid_dealloc,   // tp_dealloc
    0,                               // tp_print
    0,                               // tp_getattr
    0,                               // tp_setattr
    0,                               // tp_compare
    0,                               // tp_repr
    0,                               // tp_as_number
    0,                               // tp_as_sequence
    0,                               // tp_as_mapping
    0,                               // tp_hash 
    0,                               // tp_call
    0,                               // tp_str
    0,                               // tp_getattro
    0,                               // tp_setattro
    0,                               // tp_as_buffer
    Py_TPFLAGS_DEFAULT,              // tp_flags
    "Triangular grid",               // tp_doc
    0,                               // tp_traverse
    (inquiry)PyTriGrid_clear,        // tp_clear
    0,                               // tp_richcompare
    0,                               // tp_weaklistoffset
    0,                               // tp_iter
    0,                               // tp_iternext
    PyTriGrid_methods,               // tp_methods
    0,                               // tp_members
    0,                               // tp_getset
    0,                               // tp_base
    0,                               // tp_dict
    0,                               // tp_descr_get
    0,                               // tp_descr_set
    0,                               // tp_dictoffset
    (initproc)PyTriGrid_init,        // tp_init
    0,                               // tp_alloc
    PyTriGrid_new,                   // tp_new
  };

  static PyMethodDef module_methods[] =
  {
    {0}
  };
  
  PyMODINIT_FUNC init_mpl_tri()
  {
    if (PyType_Ready(&PyTriGridType) < 0)
    {
      return;
    }
    
    PyObject* module = Py_InitModule3("_mpl_tri", module_methods,
				      "Underlying implementation of contouring"
				      "on triangular grids.");
    if (!module)
    {
      return;
    }
    
    import_array();
    Py_INCREF(&PyTriGridType);
    PyModule_AddObject(module, "PyTriGrid", (PyObject*)&PyTriGridType);
  }
}
